
"""Contains class Population."""

import copy as cp
import pandas as pd
import numpy as np
import joblib as jl

from opqua.model import *

class Gillespie(object):
    """Class contains methods for simulating model with Gillespie algorithm.

    Class defines a model's events and methods for changing system state
    according to the possible events and simulating a timecourse using the
    Gillespie algorithm.

    Methods:
    getRates -- returns array containing rates for each event for a given system
        state
    doAction -- carries out an event, modifying system state
    run -- simulates model for a specified length of time
    """

    # Event ID constants:

    MIGRATE = 0
    CONTACT_INFECTED_HOST_ANY_HOST = 1
    CONTACT_INFECTED_HOST_ANY_VECTOR = 2
    CONTACT_HEALTHY_HOST_INFECTED_VECTOR = 3
    RECOVER_HOST = 4
    RECOVER_VECTOR = 5
    MUTATE_HOST = 6
    MUTATE_VECTOR = 7
    RECOMBINE_HOST = 8
    RECOMBINE_VECTOR = 9
    KILL_HOST = 10
    KILL_VECTOR = 11

    def __init__(self, model):
        """Create a new Gillespie simulation object.

        Arguments:
        model -- the model this simulation belongs to (Model)
        """

        super(Gillespie, self).__init__() # initialize as parent class object

        # Event IDs
        self.evt_IDs = [
            self.MIGRATE, self.CONTACT_INFECTED_HOST_ANY_HOST,
            self.CONTACT_INFECTED_HOST_ANY_VECTOR,
            self.CONTACT_HEALTHY_HOST_INFECTED_VECTOR,
            self.RECOVER_HOST, self.RECOVER_VECTOR,
            self.MUTATE_HOST, self.MUTATE_VECTOR,
            self.RECOMBINE_HOST, self.RECOMBINE_VECTOR,
            self.KILL_HOST, self.KILL_VECTOR
            ]
            # event IDs in specific order to be used

        self.model = model

    def getRates(self,population_ids):
        """Calculates event rates according to current system state.

        Arguments:
        population_ids -- list with ids for every population in the model
            (list of Strings)

        Returns:
        dictionary with event ID constants as keys and rates as values. Includes
            total rate under 'tot' key.
        """

        rates = np.zeros( [ len(self.evt_IDs), len(population_ids) ] )
            # rate array size of event space

        rates[self.MIGRATE,:] = np.array( [
            len( self.model.populations[id].hosts )
            * self.model.populations[id].total_migration_rate
            for id in population_ids
            ] )

        rates[self.CONTACT_INFECTED_HOST_ANY_HOST,:] = np.array( [
            ( len( self.model.populations[id].infected_hosts )
            / np.maximum( len( self.model.populations[id].hosts ), 1 ) )
            * self.model.populations[id].contact_rate_host_host
            for id,p in self.model.populations.items()
            ] )
            # contact rates assumes scaling area--large populations are equally
            # dense as small ones, so contact is constant with both host and
            # vector populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
            # Examines contacts between infected hosts and all hosts

        rates[self.CONTACT_INFECTED_HOST_ANY_VECTOR,:] = np.array( [
            ( len( self.model.populations[id].infected_hosts )
            / np.maximum( len( self.model.populations[id].hosts ), 1 ) )
            * self.model.populations[id].contact_rate_host_vector
            for id,p in self.model.populations.items()
            ] )
            # contact rates assumes scaling area--large populations are equally
            # dense as small ones, so contact is constant with both host and
            # vector populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
            # Examines contacts between infected hosts and all vectors

        rates[self.CONTACT_HEALTHY_HOST_INFECTED_VECTOR,:] = np.array( [
            ( len( self.model.populations[id].healthy_hosts )
            / np.maximum( len( self.model.populations[id].hosts ), 1 ) )
            * ( len( self.model.populations[id].infected_vectors )
            / np.maximum( len( self.model.populations[id].vectors ), 1 ) )
            * self.model.populations[id].contact_rate_host_vector
            for id,p in self.model.populations.items()
            ] )
            # contact rates assumes scaling area--large populations are equally
            # dense as small ones, so contact is constant with both host and
            # vector populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
            # Examines contacts between healthy hosts and infected vectors
            # (infected-infected contacts are considered in
            # CONTACT_INFECTED_HOST_VECTOR).

        rates[self.RECOVER_HOST,:] = np.array( [
            len( self.model.populations[id].infected_hosts )
            * self.model.populations[id].recovery_rate_host
            for id,p in self.model.populations.items()
            ] )

        rates[self.RECOVER_VECTOR,:] = np.array( [
            len( self.model.populations[id].infected_vectors )
            * self.model.populations[id].recovery_rate_vector
            for id,p in self.model.populations.items()
            ] )

        rates[self.MUTATE_HOST,:] = np.array( [
            len( self.model.populations[id].infected_hosts )
            * self.model.populations[id].mutate_in_host
            for id,p in self.model.populations.items()
            ] )

        rates[self.MUTATE_VECTOR,:] = np.array( [
            len( self.model.populations[id].infected_vectors )
            * self.model.populations[id].mutate_in_vector
            for id,p in self.model.populations.items()
            ] )

        rates[self.RECOMBINE_HOST,:] = np.array( [
            len( self.model.populations[id].infected_hosts )
            * self.model.populations[id].recombine_in_host
            for id,p in self.model.populations.items()
            ] )

        rates[self.RECOMBINE_VECTOR,:] = np.array( [
            len( self.model.populations[id].infected_vectors )
            * self.model.populations[id].recombine_in_vector
            for id,p in self.model.populations.items()
            ] )

        rates[self.KILL_HOST,:] = np.array( [
            len( self.model.populations[id].infected_hosts )
            * self.model.populations[id].death_rate_host
            for id,p in self.model.populations.items()
            ] )

        rates[self.KILL_VECTOR,:] = np.array( [
            len( self.model.populations[id].infected_vectors )
            * self.model.populations[id].death_rate_vector
            for id,p in self.model.populations.items()
            ] )

        return rates

    def doAction(self,act,pop,rand):

        """Change system state according to act argument passed

        Arguments:
        act -- defines action to be taken, one of the event ID constants (int)
        pop -- population action will happen in (Population)
        rand -- random number used to define event (number 0-1)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = False

        if act == self.MIGRATE:
            rand = rand * pop.total_migration_rate
            r_cum = 0
            for neighbor in pop.neighbors:
                r_cum += pop.neighbors[neighbor]
                if r_cum > rand:
                    pop.migrate(neighbor,1,0)
                    changed = True
                    break



        elif act == self.CONTACT_INFECTED_HOST_ANY_HOST:
            rand = rand * len(pop.infected_hosts)
            infected_host = int( np.floor(rand) )
            other_host = int(
                np.floor( ( rand - infected_host ) * len(pop.hosts) )
                )
            changed = pop.contactInfectedHostAnyHost(infected_host,other_host)

        elif act == self.CONTACT_INFECTED_HOST_ANY_VECTOR:
            rand = rand * len(pop.infected_hosts)
            infected_host = int( np.floor(rand) )
            vector = int(
                np.floor( ( rand - infected_host ) * len(pop.vectors) )
                )
            changed = pop.contactInfectedHostAnyVector(infected_host,vector)

        elif act == self.CONTACT_HEALTHY_HOST_INFECTED_VECTOR:
            rand = rand * len(pop.healthy_hosts)
            healthy_host = int( np.floor(rand) )
            infected_vector = int(
                np.floor( ( rand - healthy_host ) * len(pop.infected_vectors) )
                )
            changed = pop.contactHealthyHostInfectedVector(
                healthy_host,infected_vector
                )

        elif act == self.RECOVER_HOST:
            host = int( np.floor( rand * len(pop.infected_hosts) ) )
            pop.recoverHost(host)
            changed = True

        elif act == self.RECOVER_VECTOR:
            vector = int( np.floor( rand * len(pop.infected_vectors) ) )
            pop.recoverVector(vector)
            changed = True

        elif act == self.MUTATE_HOST:
            host = int( np.floor( rand * len(pop.infected_hosts) ) )
            pop.mutateHost(host)
            changed = True

        elif act == self.MUTATE_VECTOR:
            vector = int( np.floor( rand * len(pop.infected_vectors) ) )
            pop.mutateVector(vector)
            changed = True

        elif act == self.RECOMBINE_HOST:
            host = int( np.floor( rand * len(pop.infected_hosts) ) )
            pop.recombineHost(host)
            changed = True

        elif act == self.RECOMBINE_VECTOR:
            vector = int( np.floor( rand * len(pop.infected_vectors) ) )
            pop.recombineVector(vector)
            changed = True

        elif act == self.KILL_HOST:
            host = int( np.floor( rand * len(pop.infected_hosts) ) )
            pop.killHost(host)
            changed = True

        elif act == self.KILL_VECTOR:
            vector = int( np.floor( rand * len(pop.infected_vectors) ) )
            pop.killVector(vector)
            changed = True

        return changed

    def run(self,t0,tf):

        """Simulate model for a specified time between two time points.

        Simulates a time series using the Gillespie algorithm.

        Arguments:
        t0 -- initial time point to start simulation at (number)
        tf -- initial time point to end simulation at (number)

        Returns:
        dictionary containing model state history, with keys=times and
            values=Model objects with model snapshot at that time point
        """

        # Simulation variables
        t_var = t0 # keeps track of time
        history = { 0: cp.deepcopy(self.model) }
        intervention_tracker = 0
            # keeps track of what the next intervention should be
        self.model.interventions = sorted(
            self.model.interventions, key=lambda i: i.time
            )

        while t_var < tf: # repeat until t reaches end of timecourse
            population_ids = list( self.model.populations.keys() )
            r = self.getRates(population_ids) # get event rates in this state
            r_tot = np.sum(r) # sum of all rates

            # Time handling
            if r_tot > 0:
                dt = np.random.exponential( 1/r_tot ) # time until next event
                t_var += dt # add time step to main timer

                if (intervention_tracker < len(self.model.interventions)
                    and t_var
                    >= self.model.interventions[intervention_tracker].time):
                    # if there are any interventions left and if it is time
                    # to make one,
                    self.model.interventions[
                        intervention_tracker
                        ].doIntervention()
                    t_var = self.model.interventions[intervention_tracker].time
                    intervention_tracker += 1 # advance the tracker

                    population_ids = list( self.model.populations.keys() )
                    r = self.getRates(population_ids)
                        # get event rates in this state
                    r_tot = np.sum(r) # sum of all rates
                    dt = np.random.exponential( 1/r_tot )
                        # time until next event
                    t_var += dt # add time step to main timer

                # Event handling
                if t_var < tf: # if still within max time
                    u = np.random.random() * r_tot
                        # random uniform number between 0 and total rate
                    r_cum = 0 # cumulative rate
                    for e in range(r.shape[0]): # for every possible event,
                        for p in range(r.shape[1]):
                            # for every possible population,
                            r_cum += r[e,p]
                                # add this event's rate to cumulative rate
                            if u < r_cum:
                                # if random number is under cumulative rate
                                print(
                                    'Simulating time: '
                                    + str(t_var) + ', event ID: ' + str(e)
                                    )
                                changed = self.doAction(
                                    e, self.model.populations[
                                        population_ids[p]
                                        ], ( u - r_cum + r[e,p] ) / r[e,p]
                                    ) # do corresponding action,
                                      # feed in renormalized random number
                                if changed:
                                    history[t_var] = cp.deepcopy(self.model)

                                break # exit event loop

                        else: # if the inner loop wasn't broken,
                            continue # continue outer loop

                        break # otherwise, break outer loop
            else: # if at end of time course,
                if intervention_tracker < len(self.model.interventions):
                    print( 'Simulating time: ' + str(t_var), e)
                    self.model.interventions[
                        intervention_tracker
                        ].doIntervention()
                    t_var = self.model.interventions[intervention_tracker].time
                    intervention_tracker += 1 # advance the tracker
                else:
                    print( 'Simulating time: ' + str(t_var), e)
                    t_var = tf

        history[tf] = cp.deepcopy(self.model)
        history[tf].history = None

        return history




"""Contains class Vector."""

import numpy as np
import copy as cp

class Vector(object):
    """Class defines vector entities to be infected by pathogens in model.

    These can infect hosts, the main entities in the model.

    Methods:
    infectHost -- infects given host with a sample of this vector's pathogens
    recover -- removes all infections
    die -- kills this vector
    applyTreatment -- removes all infections with genotypes susceptible to given
        treatment
    """

    def __init__(self, population, id):
        """Create a new Vector.

        Arguments:
        population -- the population this vector belongs to (Population)
        id -- unique identifier for this vector within population (String)
        """
        super(Vector, self).__init__()
        self.id = id
        self.pathogens = {} # Dictionary with all current infections in this
            # vector, with keys=genome strings, values=fitness numbers
        self.population = population
        self.sum_fitness = 0 # sum of all pathogen fitnesses within this vector
        self.protection_sequences = [] # A list of strings this vector is immune
            # to. If a pathogen's genome contains one of these values, it cannot
            # infect this vector.

    def infectHost(self, host):
        """Infect given host with a sample of this vector's pathogens.

        Each pathogen in the infector is sampled as present or absent in the
        inoculum by drawing from a Poisson distribution with a mean equal to the
        mean inoculum size of the organism being infected weighed by each
        genome's fitness as a fraction of the total in the infector) as the
        probability of each trial. Each pathogen present in the inoculum will be
        added to the infected organism, if it does not have protection from the
        pathogen's genome. Fitnesses are computed for the pathogens' genomes in
        the infected organism, and the organism is included in the poplation's
        infected list if appropriate.

        Arguments:
        host -- the host to be infected (Host)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = False
        for genome,fitness in self.pathogens.items():
            if ( genome not in host.pathogens.keys()
                    and genome not in host.protection_sequences
                    and np.random.poisson(
                        self.population.mean_inoculum_host
                            * fitness / self.sum_fitness, 1
                        )
                    > 0 ):
                host.pathogens[genome] = host.population.fitnessHost(genome)
                host.sum_fitness += host.pathogens[genome]
                changed = True
                if host not in host.population.infected_hosts:
                    host.population.infected_hosts.append(host)
                    host.population.healthy_hosts.remove(host)

        return changed

    def recover(self):
        """Remove all infections from this vector.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.
        """

        if self.population.protection_upon_recovery_host:
            for genome in self.pathogens:
                seq = genome[self.protection_upon_recovery_host[0]
                    :self.protection_upon_recovery_host[1]]
                if seq not in self.protection_sequences:
                    self.protection_sequences.append(seq)

        self.pathogens = {}
        if self in self.population.infected_vectors:
            self.population.infected_vectors.remove(self)

    def die(self):
        """Add vector to population's dead list, remove it from alive ones."""

        self.population.dead_vectors.append(self)
        self.population.vectors.remove(self)

    def applyTreatment(self, resistance_seqs):
        """Remove all infections with genotypes susceptible to given treatment.

        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)
        """

        genomes_removed = []
        for genome in self.pathogens:
            for seq in resistance_seqs:
                if seq not in genome:
                    genomes_removed.append(genome)
                    break

        for g in genomes_removed:
            del self.pathogens[g]

        if ( len(self.pathogens) == 0
                and self in self.population.infected_vectors ):
            self.population.infected_vectors.remove(self)




"""Contains class Population."""

import numpy as np
import copy as cp

from opqua.internal.host import Host
from opqua.internal.vector import Vector

class Population(object):
    """Class defines a population with hosts, vectors, and specific parameters.

    Methods:
    setSetup -- assigns a given set of parameters to this population
    addHosts -- adds hosts to the population
    addVectors -- adds vectors to the population
    newHostGroup -- returns a list of random (healthy or any) hosts
    newVectorGroup -- returns a list of random (healthy or any) vectors
    removeHosts -- removes hosts from the population
    removeVectors -- removes vectors from the population
    addPathogensToHosts -- adds pathogens with specified genomes to hosts
    addPathogensToVectors -- adds pathogens with specified genomes to vectors
    recoverHost --removes all infections from given host
    recoverVector -- removes all infections from given vector
    treatHosts -- removes infections susceptible to given treatment from hosts
    treatVectors -- removes infections susceptible to treatment from vectors
    protectHosts -- adds protection sequence to hosts
    protectVectors -- adds protection sequence to vectors
    setNeighbor -- sets migration rate from this population towards another
    migrate -- transfers hosts and/or vectors from this population to a neighbor
    contactInfectedHostAnyHost -- carries out a contact event between a random
        infected host and any random host in population
    contactInfectedHostAnyVector -- carries out a contact event between a random
        infected host and any random vector in population
    contactHealthyHostInfectedVector -- carries out a contact event between a
        random healthy host and a random infected vector in population
    mutateHost -- mutates a single locus in a random pathogen in a host
    mutateVector -- mutates a single locus in a random pathogen in a vector
    recombineHost -- recombines two random pathogens in a host
    recombineVector -- recombines two random pathogens in a host
    """

    def __init__(self, model, id, setup, num_hosts, num_vectors):
        """Create a new Population.

        Arguments:
        model -- the model this population belongs to (Model)
        id -- unique identifier for this population in the model (String)
        setup -- setup object with parameters for this population (Setup)
        num_hosts -- number of hosts to initialize population with (int)
        num_vectors -- number of hosts to initialize population with (int)
        """
        super(Population, self).__init__()

        self.id = id

        self.hosts = [
            Host( self, self.id + '_' + str(id) ) for id in range(num_hosts)
            ]
            # contains all live hosts
        self.vectors = [
            Vector(self, self.id + '_' + str(id) ) for id in range(num_vectors)
            ]
            # contains all live vectors
        self.infected_hosts = []
            # contains live, infected hosts (also in hosts list)
        self.healthy_hosts = self.hosts.copy()
            # contains live, healthy hosts (also in hosts list)
        self.infected_vectors = []
            # contains live, infected vectors (also in vectors list)
        self.dead_hosts = [] # contains dead hosts (not in hosts list)
        self.dead_vectors = [] # contains dead vectors (not in vectors list)
        self.neighbors = {} # dictionary with neighboring populations,
            # keys=Population, values=migration rate from this population to
            # neighboring population

        self.total_migration_rate = 0 # sum of all migration rates from this
            # population to neighbors

        self.setSetup(setup)

    def setSetup(self, setup):
        """Assign parameters stored in Setup object to this population.

        Arguments:
        setup -- the setup to be assigned (Setup)
        """

        self.setup = setup

        self.num_loci = setup.num_loci
        self.possible_alleles = setup.possible_alleles

        self.fitnessHost = setup.fitnessHost
        self.fitnessVector = setup.fitnessVector
        self.contact_rate_host_vector = setup.contact_rate_host_vector
        self.contact_rate_host_host = setup.contact_rate_host_host
            # contact rate assumes fixed area--large populations are dense
            # populations, so contact scales linearly with both host and vector
            # populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
        self.mean_inoculum_host = setup.mean_inoculum_host
        self.mean_inoculum_vector = setup.mean_inoculum_vector
        self.recovery_rate_host = setup.recovery_rate_host
        self.recovery_rate_vector = setup.recovery_rate_vector
        self.recombine_in_host = setup.recombine_in_host
        self.recombine_in_vector = setup.recombine_in_vector
        self.mutate_in_host = setup.mutate_in_host
        self.mutate_in_vector = setup.mutate_in_vector
        self.death_rate_host = setup.death_rate_host
        self.death_rate_vector = setup.death_rate_vector
        self.protection_upon_recovery_host \
            = setup.protection_upon_recovery_host
        self.protection_upon_recovery_vector \
            = setup.protection_upon_recovery_vector

    def addHosts(self, num_hosts):
        """Add a number of healthy hosts to population, return list with them.

        Arguments:
        num_hosts -- number of hosts to be added (int)

        Returns:
        list containing new hosts
        """

        new_hosts = [
            Host(
                self, self.id + '_' + str( i + len(self.hosts) )
                ) for i in range(num_hosts)
            ]
        self.hosts += new_hosts
        self.healthy_hosts += new_hosts

        return new_hosts

    def addVectors(self, num_vectors):
        """Add a number of healthy vectors to population, return list with them.

        Arguments:
        num_vectors -- number of vectors to be added (int)

        Returns:
        list containing new vectors
        """

        new_vectors = [
            Vector(
                self, self.id + '_' + str( i + len(self.vectors) )
                ) for i in range(num_vectors)
            ]
        self.vectors += new_vectors

        return new_vectors

    def newHostGroup(self, num_hosts, healthy=True):
        """Return a list of random (healthy or any) hosts in population.

        Arguments:
        num_vectors -- number of vectors to be sampled randomly (int)

        Keyword arguments:
        healthy -- whether to sample healthy hosts only (default True; Boolean)

        Returns:
        list containing sampled hosts
        """

        possible_hosts = []
        if healthy:
            if len(self.healthy_hosts) >= num_hosts:
                possible_hosts = self.healthy_hosts
            else:
                raise ValueError(
                    "You're asking for " + str(num_hosts)
                    + " healthy hosts, but population " + str(self.id)
                    + " only has " + str( len(self.healthy_hosts) ) + "."
                    )
        else:
            possible_hosts = self.hosts

        hosts = np.random.choice(possible_hosts, num_hosts, replace=False)

        return hosts

    def newVectorGroup(self, num_vectors, healthy=True):
        """Return a list of random (healthy or any) vectors in population.

        Arguments:
        num_vectors -- number of vectors to be sampled randomly (int)

        Keyword arguments:
        healthy -- whether to sample healthy vectors only (default True;
            Boolean)

        Returns:
        list containing sampled vectors
        """

        possible_vectors = []
        if healthy:
            if len(self.vectors) - len(self.infected_vectors) >= num_vectors:
                for vector in self.vectors:
                    if vector not in self.infected_vectors:
                        possible_vectors.append(vector)
            else:
                raise ValueError(
                    "You're asking for " + str(num_vectors)
                    + " healthy vectors, but population " + str(self.id)
                    + " only has "
                    + str( len(self.vectors) - len(self.infected_vectors) )
                    + "."
                    )
        else:
            possible_vectors = self.vectors

        vectors = np.random.choice(possible_vectors, num_vectors, replace=False)

        return vectors

    def removeHosts(self, num_hosts_or_list):
        """Remove a number of specified or random hosts from population.

        Arguments:
        num_hosts_or_list -- number of hosts to be sampled randomly for removal
            or list of hosts to be removed, must be hosts in this population
            (int or list of Hosts)
        """

        if isinstance(num_hosts_or_list, list):
            for host_removed in num_hosts_or_list:
                if host_removed in self.hosts:
                    if host_removed in self.infected_hosts:
                        self.infected_hosts.remove( host_removed )
                    else:
                        self.healthy_hosts.remove( host_removed )

                    self.hosts.remove( host_removed )
        else:
            for _ in range(num_hosts_or_list):
                host_removed = np.random.choice(self.hosts)
                if host_removed in self.infected_hosts:
                    self.infected_hosts.remove( host_removed )
                else:
                    self.healthy_hosts.remove( host_removed )

                self.hosts.remove( host_removed )

    def removeVectors(self, num_vectors_or_list):
        """Remove a number of specified or random vectors from population.

        Arguments:
        num_vectors_or_list -- number of vectors to be sampled randomly for
            removal or list of vectors to be removed, must be vectors in this
            population (int or list of Vectors)
        """

        if isinstance(num_vectors_or_list, list):
            for vector_removed in num_vectors_or_list:
                if vector_removed in self.vectors:
                    if vector_removed in self.infected_vectors:
                        self.infected_vectors.remove( vector_removed )

                    self.vectors.remove( vector_removed )
        else:
            for _ in range(num_vectors):
                vector_removed = np.random.choice(self.vectors)
                if vector_removed in self.infected_vectors:
                    self.infected_vectors.remove( vector_removed )

                self.vectors.remove( vector_removed )

    def addPathogensToHosts(self, genomes_numbers, hosts=[]):
        """Add specified pathogens to random hosts, optionally from a list.

        Arguments:
        genomes_numbers -- dictionary conatining pathogen genomes to add as keys
            and number of hosts each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        for genome in genomes_numbers:
            if len(genome) == self.num_loci and all( [
                allele in self.possible_alleles[i]
                    for i,allele in enumerate(genome)
                ] ):
                new_fitness = self.fitnessHost(genome)
                if len(hosts) == 0:
                    hosts = self.hosts

                rand_hosts = np.random.choice(
                    hosts, genomes_numbers[genome], replace=False
                    )
                for rand_host in rand_hosts:
                    if rand_host not in self.infected_hosts:
                        self.infected_hosts.append(rand_host)
                        self.healthy_hosts.remove(rand_host)

                    if genome not in rand_host.pathogens:
                        rand_host.pathogens[genome] = new_fitness
                        rand_host.sum_fitness += new_fitness
            else:
                raise ValueError('Genome ' + genome + ' must be of length '
                    + str(self.num_loci)
                    + ' and contain only the following characters at each '
                    + 'position: ' + self.possible_alleles + ' .')

    def addPathogensToVectors(self, genomes_numbers, vectors=[]):
        """Add specified pathogens to random vectors, optionally from a list.

        Arguments:
        genomes_numbers -- dictionary conatining pathogen genomes to add as keys
            and number of vectors each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        for genome in genomes_numbers:
            if len(genome) == self.num_loci and all( [
                allele in self.possible_alleles[i]
                    for i,allele in enumerate(genome)
                ] ):
                new_fitness = self.fitnessVector(genome)
                if len(vectors) == 0:
                    vectors = self.vectors

                rand_vectors = np.random.choice(
                    vectors, genomes_numbers[genome], replace=False
                    )
                for rand_vector in rand_vectors:
                    if rand_vector not in self.infected_vectors:
                        self.infected_vectors.append(rand_vector)

                    if genome not in rand_vector.pathogens:
                        rand_vector.pathogens[genome] = new_fitness
                        rand_vector.sum_fitness += new_fitness
            else:
                raise ValueError('Genome ' + genome + ' must be of length '
                    + str(self.num_loci)
                    + ' and contain only the following characters at each '
                    + 'position: ' + self.possible_alleles + ' .')

    def recoverHost(self, index_host):
        """Remove all infections from host at this index.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.

        Arguments:
        index_host -- index of host in infected_hosts (int)
        """

        self.infected_hosts[index_host].recover()

    def recoverVector(self, index_vector):
        """Remove all infections from vector at this index.

        If model is protecting upon recovery, add protecion sequence as defined
        by the indexes in the corresponding model parameter. Remove from
        population infected list and add to healthy list.

        Arguments:
        index_vector -- index of vector in infected_vectors (int)
        """

        self.infected_vectors[index_vector].recover()

    def killHost(self, index_host):
        """Add host at this index to dead list, remove it from alive ones.

        Arguments:
        index_host -- index of host in infected_hosts (int)
        """

        self.infected_hosts[index_host].die()

    def killVector(self, index_vector):
        """Add host at this index to dead list, remove it from alive ones.

        Arguments:
        index_vector -- index of vector in infected_vectors (int)
        """

        self.infected_vectors[index_vector].die()

    def treatHosts(self, frac_hosts, resistance_seqs, hosts=[]):
        """Treat random fraction of infected hosts against some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        hosts_to_consider = self.hosts
        if len(hosts) > 0:
            hosts_to_consider = hosts

        possible_infected_hosts = []
        for host in hosts_to_consider:
            if len( host.pathogens ):
                possible_infected_hosts.append( host )

        treat_hosts = np.random.choice(
            possible_infected_hosts,
            int( frac_hosts * len( possible_infected_hosts ) ), replace=False
            )
        for host in treat_hosts:
            possible_infected_hosts[host].applyTreatment(resistance_seqs)

    def treatVectors(self, frac_vectors, resistance_seqs, vectors=[]):
        """Treat random fraction of infected vectors agains some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        vectors_to_consider = self.vectors
        if len(vectors) > 0:
            vectors_to_consider = vectors

        possible_infected_vectors = []
        for vector in vectors_to_consider:
            if len( vector.pathogens ):
                possible_infected_vectors.append( vector )

        treat_vectors = np.random.choice(
            possible_infected_vectors,
            int( frac_vectors * len( possible_infected_vectors ) ),
            replace=False
            )
        for vector in treat_vectors:
            possible_infected_vectors[vector].applyTreatment(resistance_seqs)

    def protectHosts(self, frac_hosts, protection_sequence, hosts=[]):
        """Protect a random fraction of infected hosts against some infection.

        Adds protection sequence specified to a random fraction of the hosts
        specified. Does not cure them if they are already infected.

        Arguments:
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        hosts_to_consider = self.hosts
        if len(hosts) > 0:
            hosts_to_consider = hosts

        protect_hosts = np.random.choice(
            self.hosts, int( frac_hosts * len( hosts_to_consider ) ),
            replace=False
            )
        for host in protect_hosts:
            host.protection_sequences.append(protection_sequence)

    def protectVectors(self, frac_vectors, protection_sequence, vectors=[]):
        """Protect a random fraction of infected vectors against some infection.

        Adds protection sequence specified to a random fraction of the vectors
        specified. Does not cure them if they are already infected.

        Arguments:
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        vectors_to_consider = self.vectors
        if len(vectors) > 0:
            vectors_to_consider = vectors

        protect_vectors = np.random.choice(
            self.vectors, int( frac_vectors * len( vectors_to_consider ) ),
            replace=False
            )
        for vector in protect_vectors:
            vector.protection_sequences.append(protection_sequence)

    def setNeighbor(self, neighbor, rate):
        """Set migration rate from this population towards another one.

         Arguments:
         neighbor -- population towards which migration rate will be specified
            (Population)
         rate -- migration rate from this population to the neighbor (number)
         """

        if neighbor in self.neighbors:
            self.total_migration_rate -= self.neighbors[neighbor]

        self.neighbors[neighbor] = rate
        self.total_migration_rate += rate

    def migrate(self, target_pop, num_hosts, num_vectors):
        """Transfer hosts and/or vectors from this population to another.

        Arguments:
        target_pop -- population towards which migration will occur (Population)
        num_hosts -- number of hosts to transfer (int)
        num_vectors -- number of vectors to transfer (int)
        """

        migrating_hosts = np.random.choice(self.hosts, num_hosts, replace=False)
        for host in migrating_hosts:
            self.hosts.remove(host)
            target_pop.hosts.append(host)
            host.population = target_pop
            if host in self.infected_hosts:
                self.infected_hosts.remove(host)
                target_pop.infected_hosts.append(host)
            else:
                self.healthy_hosts.remove(host)
                target_pop.healthy_hosts.append(host)

        migrating_vectors = np.random.choice(
            self.vectors,num_vectors, replace=False
            )
        for vector in migrating_vectors:
            self.vectors.remove(vector)
            target_pop.vectors.append(vector)
            vector.population = target_pop
            if vector in self.infected_vectors:
                self.infected_vectors.remove(vector)
                target_pop.infected_vectors.append(vector)

    def contactInfectedHostAnyHost(self, index_infected_host, index_other_host):
        """Contact a random infected host and any random host in population.

        Carries out possible infection events from each organism into the other.

        Arguments:
        index_infected_host -- index of infected host in infected_hosts (int)
        index_other_host -- index of second host in hosts (int)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        temp_host = cp.deepcopy(self.infected_hosts[index_infected_host])
        changed1 = self.hosts[index_other_host].infectHost(
            self.infected_hosts[index_infected_host]
            )
        changed2 = temp_host.infectHost(self.hosts[index_other_host])

        return ( changed1 or changed2 )

    def contactInfectedHostAnyVector(self, index_host, index_vector):
        """Contact a random infected host and any random vector in population.

        Carries out possible infection events from each organism into the other.

        Arguments:
        index_host -- index of infected host in infected_hosts (int)
        index_vector -- index of vector in vectors (int)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        temp_host = cp.deepcopy(self.infected_hosts[index_host])
        changed1 = self.vectors[index_vector].infectHost(
            self.infected_hosts[index_host]
            )
        changed2 = temp_host.infectVector(self.vectors[index_vector])

        return ( changed1 or changed2 )

    def contactHealthyHostInfectedVector(self, index_host, index_vector):
        """Contact a random healthy host and a random infected vector.

        Carries out possible infection events from vector into host.

        Arguments:
        index_host -- index of host in healthy_hosts (int)
        index_vector -- index of infected vector in infected_vectors (int)

        Returns:
        whether or not the model has changed state (Boolean)
        """

        changed = self.infected_vectors[index_vector].infectHost(
            self.healthy_hosts[index_host]
            )

        return changed

    def mutateHost(self, index_host):
        """Mutate a single, random locus in a random pathogen in the given host.

        Creates a new genotype from a de novo mutation event in the host given.

        Arguments:
        index_host -- index of host in infected_hosts (int)
        """

        host = self.infected_hosts[index_host]

        if len(host.pathogens) > 0:
            old_genome = np.random.choice( list( host.pathogens.keys() ) )
            mut_index = np.random.randint( self.num_loci )
            new_genome = old_genome[0:mut_index] + np.random.choice(
                list(self.possible_alleles[mut_index])
                ) + old_genome[mut_index+1:]
            host.pathogens[new_genome] = self.fitnessHost(new_genome)
            host.sum_fitness += host.pathogens[new_genome]

    def mutateVector(self, index_vector):
        """Mutate a single, random locus in a random pathogen in given vector.

        Creates a new genotype from a de novo mutation event in the vector
        given.

        Arguments:
        index_vector -- index of vector in infected_vectors (int)
        """

        vector = self.infected_vectors[index_vector]

        if len(vector.pathogens) > 0:
            old_genome = np.random.choice( list( vector.pathogens.keys() ) )
            mut_index = np.random.randint( self.num_loci )
            new_genome = old_genome[0:mut_index] + np.random.choice(
                list(self.possible_alleles[mut_index])
                ) + old_genome[mut_index+1:]
            vector.pathogens[new_genome] = self.fitnessHost(new_genome)
            vector.sum_fitness += vector.pathogens[new_genome]

    def recombineHost(self, index_host):
        """Recombine two random pathogen genomes at random locus in given host.

        Creates a new genotype from two random possible pathogens in the host
        given.

        Arguments:
        index_host -- index of host in infected_hosts (int)
        """

        host = self.infected_hosts[index_host]

        if len(host.pathogens) > 1:
            parents = np.random.choice( list( host.pathogens.keys() ), 2,
                replace=False )
            recom_index = np.random.randint( self.num_loci )
            new_genome = parents[0][0:recom_index] + parents[1][recom_index:]
            host.pathogens[new_genome] = self.fitnessHost(new_genome)
            host.sum_fitness += host.pathogens[new_genome]

    def recombineVector(self, index_vector):
        """Recombine 2 random pathogen genomes at random locus in given vector.

        Creates a new genotype from two random possible pathogens in the vector
        given.

        Arguments:
        index_vector -- index of vector in infected_vectors (int)
        """

        vector = self.infected_vectors[index_vector]

        if len(vector.pathogens) > 1:
            parents = np.random.choice( list( vector.pathogens.keys() ), 2,
                replace=False )
            recom_index = np.random.randint( self.num_loci )
            new_genome = parents[0][0:recom_index] + parents[1][recom_index:]
            vector.pathogens[new_genome] = self.fitnessHost(new_genome)
            vector.sum_fitness += vector.pathogens[new_genome]




"""Contains data wrangling methods."""

import numpy as np # handle arrays
import pandas as pd # data wrangling
import copy as cp
import joblib as jl

def saveToDf(history,save_to_file,n_cores=0):
    """Save status of model to dataframe, write to file location given.

    Creates a pandas Dataframe in long format with the given model history, with
    one host or vector per simulation time in each row, and columns:
        Time - simulation time of entry
        Population - ID of this host/vector's population
        Organism - host/vector
        ID - ID of host/vector
        Pathogens - all genomes present in this host/vector separated by ;
        Protection - all genomes present in this host/vector separated by ;
        Alive - whether host/vector is alive at this time, True/False

    Writing straight to a file and then reading into a pandas dataframe was
    actually more efficient than concatenating directly into a pd dataframe.

    Arguments:
    history -- dictionary containing model state history, with keys=times and
        values=Model objects with model snapshot at that time point
    save_to_file -- file path and name to save model data under (String)

    Keyword arguments:
    n_cores -- number of cores to parallelize file export across, if 0, all
        cores available are used (default 0; int)

    Returns:
    pandas dataframe with model history as described above
    """

    print('Saving file...')

    if not n_cores:
        n_cores = jl.cpu_count()

    new_df = ','.join(
        ['Time','Population','Organism','ID','Pathogens','Protection','Alive']
        ) + '\n' + '\n'.join( jl.Parallel(n_jobs=n_cores, verbose=10) (
            jl.delayed( lambda d: ''.join(d) ) (
                '\n'.join( [
                    '\n'.join( [ ','.join( [
                        str(time), str(pop.id), 'Host', str(host.id), '"'
                        + ';'.join( host.pathogens.keys() )
                        + '"', '"' + ';'.join( host.protection_sequences )
                        + '"', 'True'
                        ] ) for host in pop.hosts ] ) + '\n'
                    + '\n'.join( [ ','.join( [
                        str(time), str(pop.id), 'Vector', str(vector.id), '"'
                        + ';'.join( vector.pathogens.keys() )
                        + '"', '"' + ';'.join( vector.protection_sequences )
                        + '"', 'True'
                        ] ) for vector in pop.vectors ] ) + '\n'
                    + '\n'.join( [ ','.join( [
                        str(time), str(pop.id), 'Host', str(host.id), '"'
                        + ';'.join( host.pathogens.keys() ) + '"', '"'
                        + ';'.join( host.protection_sequences ) + '"', 'False'
                        ] ) for host in pop.dead_hosts ] ) + '\n'
                    + '\n'.join( [ ','.join( [
                        str(time), str(pop.id), 'Vector', str(vector.id), '"'
                        + ';'.join( vector.pathogens.keys() ) + '"', '"'
                        + ';'.join( vector.protection_sequences ) + '"', 'False'
                        ] ) for vector in pop.dead_vectors ] )
                    for id,pop in model.populations.items()
                ] )
            ) for time,model in history.items()
        ) )

    new_df = new_df.replace(
        '\n\n','\n'
        ).replace('\n\n','\n').replace('\n\n','\n')

    file = open(save_to_file,'w')
    file.write(new_df)
    file.close()

    new_df = pd.read_csv(save_to_file)

    print('...file saved.')

    return new_df

def populationsDf(
        data, compartment='Infected', hosts=True, vectors=False,
        num_top_populations=-1, track_specific_populations=[], save_to_file=""):
    """Create dataframe with aggregated totals per population.

    Creates a pandas Dataframe in long format with dynamics of a compartment
    across populations in the model, with one time point in each row and columns
    for time as well as each population.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    compartment -- subset of hosts/vectors to count totals of, can be either
        'Naive','Infected','Recovered', or 'Dead' (default 'Infected'; String)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    num_top_populations -- how many populations to count separately and include
        as columns, remainder will be counted under column "Other"; if <0,
        includes all populations in model (default -1; int)
    track_specific_populations -- contains IDs of specific populations to have
        as a separate column if not part of the top num_top_populations
        populations (list of Strings)
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)

    Returns:
    pandas dataframe with model population dynamics as described above
    """

    dat = cp.deepcopy( data )

    if not hosts:
        dat = dat[ dat['Organism'] != 'Host' ]
    if not vectors:
        dat = dat[ dat['Organism'] != 'Vector' ]

    if num_top_populations < 0:
        num_top_populations = len( pd.unique( dat['Population'] ) )

    dat['Infected'] = ( dat['Pathogens'].fillna('').str.len() > 0 )
    dat['Protected'] = ( dat['Protection'].fillna('').str.len() > 0 )

    grouped = dat.groupby( [
        'Time','Population','Alive','Infected','Protected'
        ] ).size().reset_index(name='Number')

    compartment_names = ['Naive','Infected','Recovered','Dead']

    grouped['Compartment'] = compartment_names[3]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == False )
            & ( grouped['Protected'] == False ), 'Compartment'
        ] = compartment_names[0]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == True ),
        'Compartment'
        ] = compartment_names[1]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == False )
        & ( grouped['Protected'] == True ), 'Compartment'
        ] = compartment_names[2]

    grouped = grouped[ grouped['Compartment'] == compartment ]
    grouped = grouped.drop(
        columns=['Alive','Infected','Protected','Compartment']
        )
    grouped = grouped.pivot(
        columns='Population', values='Number', index='Time'
        ).fillna(0).reset_index('Time')

    for pop in pd.unique(data['Population']):
        if pop not in grouped.columns:
            grouped[pop] = 0

    if len(grouped.columns)-1 < num_top_populations:
        num_top_populations = len(grouped.columns)-1

    populations_to_drop = list(grouped.columns)[ num_top_populations+1: ]
    for pop in track_specific_populations:
        if pop in populations_to_drop:
            populations_to_drop.remove(pop)

    grouped['Other'] = 0
    if len(populations_to_drop) > 0:
        grouped['Other'] = grouped[populations_to_drop].sum(axis=1)

    if grouped['Other'].sum() == 0:
        populations_to_drop = populations_to_drop + ['Other']

    grouped = grouped.drop( columns=populations_to_drop )

    if len(save_to_file) > 0:
        grouped.to_csv(save_to_file, index=False)

    return grouped


def compartmentDf(
        data, populations=[], hosts=True, vectors=False, save_to_file=""):
    """Create dataframe with number of naive, susc., inf., rec. hosts/vectors.

    Creates a pandas Dataframe with dynamics of all compartments (naive,
    infected, recovered, dead) across selected populations in the model,
    with one time point in each row and columns for time as well as each
    compartment.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    populations -- IDs of populations to include in analysis; if empty, uses all
        populations in model (default empty list; list of Strings)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)

    Returns:
    pandas dataframe with model compartment dynamics as described above
    """

    if len(populations) > 0:
        dat = cp.deepcopy( data[ data['Population'].isin( populations ) ] )
    else:
        dat = cp.deepcopy( data )

    if not hosts:
        dat = dat[ dat['Organism'] != 'Host' ]

    if not vectors:
        dat = dat[ dat['Organism'] != 'Vector' ]

    dat['Infected'] = ( dat['Pathogens'].fillna('').str.len() > 0 )
    dat['Protected'] = ( dat['Protection'].fillna('').str.len() > 0 )

    grouped = dat.groupby( [
        'Time','Alive','Infected','Protected'
        ] ).size().reset_index(name='Number')

    compartment_names = ['Naive','Infected','Recovered','Dead']

    grouped['Compartment'] = compartment_names[3]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == False )
        & ( grouped['Protected'] == False ), 'Compartment'
        ] = compartment_names[0]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == True )
        & ( grouped['Protected'] == True ), 'Compartment'
        ] = compartment_names[1]
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == True )
        & ( grouped['Protected'] == False ), 'Compartment'
        ] = compartment_names[1] + '_2'
    grouped.loc[
        ( grouped['Alive'] == True ) & ( grouped['Infected'] == False )
        & ( grouped['Protected'] == True ), 'Compartment'
        ] = compartment_names[2]

    grouped = grouped.drop( columns=['Alive','Infected','Protected'] )
    grouped = grouped.pivot(
        columns='Compartment', values='Number', index='Time'
        ).fillna(0).reset_index('Time')

    if ( compartment_names[1] in grouped.columns
            and compartment_names[1] + '_2' in grouped.columns ):
        grouped[compartment_names[1]] = ( grouped[ compartment_names[1] ]
            + grouped[ compartment_names[1] + '_2' ] )
        grouped = grouped.drop( columns=[ compartment_names[1] + '_2' ])
    elif ( compartment_names[1] + '_2' in grouped.columns ):
        grouped[compartment_names[1]] = grouped[ compartment_names[1] + '_2' ]
        grouped = grouped.drop( columns=[ compartment_names[1] + '_2' ])

    for comp_name in compartment_names:
        if comp_name not in grouped.columns:
            grouped[comp_name] = 0


    if len(save_to_file) > 0:
        grouped.to_csv(save_to_file, index=False)

    return grouped


def compositionDf(
        data, populations=[], type_of_composition='Pathogens', hosts=True,
        vectors=False, num_top_sequences=-1, track_specific_genomes=[],
        save_to_file=""):
    """Create dataframe with counts for pathogen genomes or resistance.

    Creates a pandas Dataframe with dynamics of the pathogen strains or
    protection sequences across selected populations in the model,
    with one time point in each row and columns for pathogen genomes or
    protection sequences.

    Of note: sum of totals for all sequences in one time point does not
    necessarily equal the number of infected hosts and/or vectors, given
    multiple infections in the same host/vector are counted separately.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    populations -- IDs of populations to include in analysis; if empty, uses all
        populations in model (default empty list; list of Strings)
    type_of_composition -- field of data to count totals of, can be either
        'Pathogens' or 'Protection' (default 'Pathogens'; String)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    num_top_sequences -- how many sequences to count separately and include
        as columns, remainder will be counted under column "Other"; if <0,
        includes all genomes in model (default -1; int)
    track_specific_genomes -- contains specific sequences to have
        as a separate column if not part of the top num_top_sequences
        sequences (list of Strings)
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)

    Returns:
    pandas dataframe with model sequence composition dynamics as described above
    """

    if len(populations) > 0:
        dat = cp.deepcopy( data[ data['Population'].isin( populations ) ] )
    else:
        dat = cp.deepcopy( data )

    if not hosts:
        dat = dat[ dat['Organism'] != 'Host' ]
    if not vectors:
        dat = dat[ dat['Organism'] != 'Vector' ]

    if num_top_sequences < 0:
        num_top_sequences = len( pd.unique( dat[type_of_composition] ) )

    all_genomes = pd.Series(
        ';'.join( dat[type_of_composition].dropna() ).split(';')
        ).str.strip()
    top_genomes = all_genomes.value_counts(ascending=False)

    if len(top_genomes) < num_top_sequences:
        num_top_sequences = len(top_genomes)

    genomes_to_track = list(top_genomes[0:num_top_sequences].index)
    for genome in track_specific_genomes:
        if genome not in genomes_to_track:
            genomes_to_track.append(genome)


    genome_split_data = []
    other_genome_data = pd.DataFrame(
        np.zeros( len( pd.unique( data['Time'] ) ) ),
        index=pd.unique( data['Time'] ), columns=['Other']
        )
    c = 0

    if len( ''.join(top_genomes.index) ) > 0:
        for genome in top_genomes.index:
            dat_genome = dat[ dat[type_of_composition].str.contains(
                genome, na=False
                ) ]
            grouped = dat_genome.groupby('Time').size().reset_index(name=genome)
            grouped = grouped.set_index('Time')

            if genome in genomes_to_track:
                genome_split_data.append(grouped)
            else:
                other_genome_data = other_genome_data.join(
                    grouped, how='outer'
                    ).fillna(0)
                other_genome_data['Other'] = ( other_genome_data['Other']
                    + other_genome_data[genome] )
                other_genome_data = other_genome_data.drop(columns=[genome])

            c += 1
            print(
                str(c) + ' / ' + str( len(top_genomes.index) )
                + ' genotypes processed.'
                )


        if other_genome_data['Other'].sum() > 0:
            genome_split_data += [other_genome_data]
            genomes_to_track += ['Other']


    times = pd.DataFrame(
        np.zeros( len( pd.unique( data['Time'] ) ) ),
        index=pd.unique( data['Time'] ), columns=['*None*']
        )
    composition = times.join( genome_split_data, how='outer' )
    composition = composition.drop(columns=['*None*']).fillna(0)
    composition = composition.reset_index()
    composition.columns = ['Time'] + list( composition.columns )[1:]

    if len(save_to_file) > 0:
        composition.to_csv(save_to_file, index=False)

    return composition


def getPathogens(data, save_to_file=""):
    """Create Dataframe with counts for all pathogen genomes in data.

    Returns sorted pandas Dataframe with counts for occurrences of all pathogen
    genomes in data passed.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)

    Returns:
    pandas dataframe with Series as described above
    """

    out = pd.Series( ';'.join(
        data['Pathogens'].dropna()
        ).split(';') ).str.strip().value_counts(ascending=False).reset_index()
    out.columns = ['Pathogens'] + list( out.columns )[1:]

    if len(save_to_file) > 0:
        out.to_csv(save_to_file, index=False)

    return out

def getProtections(data, save_to_file=""):
    """Create Dataframe with counts for all protection sequences in data.

    Returns sorted pandas Dataframe with counts for occurrences of all
    protection sequences in data passed.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)

    Returns:
    pandas dataframe with Series as described above
    """

    out = pd.Series( ';'.join(
        data['Protection'].dropna()
        ).split(';') ).str.strip().value_counts(ascending=False).reset_index()
    out.columns = ['Protection'] + list( out.columns )[1:]

    if len(save_to_file) > 0:
        out.to_csv(save_to_file, index=False)

    return out



# TODO: Update all fitnesses
"""Contains class Model; main class user interacts with."""

import numpy as np
import pandas as pd
import difflib as dl

from opqua.internal.host import Host
from opqua.internal.vector import Vector
from opqua.internal.population import Population
from opqua.internal.setup import Setup
from opqua.internal.intervention import Intervention
from opqua.internal.gillespie import Gillespie
from opqua.internal.data import saveToDf
from opqua.internal.plot import populationsPlot, compartmentPlot, \
    compositionPlot

class Model(object):
    """Class defines a Model.

    This is the main class that the user interacts with.

    The Model class contains populations, setups, and interventions to be used
    in simulation. Also contains groups of hosts/vectors for manipulations and
    stores model history as snapshots for each time point.

    *** --- ATTRIBUTES: --- ***
    populations -- dictionary with keys=population IDs, values=Population
        objects
    setups -- dictionary with keys=setup IDs, values=Setup objects
    interventions -- contains model interventions in the order they will occur
    groups -- dictionary with keys=group IDs, values=lists of hosts/vectors
    self.history -- dictionary with keys=time values, values=Model objects that
        are snapshots of Model at that timepoint

    *** --- METHODS: --- ***

    --- Model Initialization and Simulation: ---

    newSetup -- creates a new Setup, save it in setups dict under given name
    newIntervention -- creates a new intervention executed during simulation
    run -- simulates model for a specified length of time

    --- Data Output and Plotting: ---

    saveToDataFrame -- saves status of model to dataframe, writes to file
    getPathogens -- creates Dataframe with counts for all pathogen genomes
    getProtections -- creates Dataframe with counts for all protection sequences
    populationsPlot -- plots aggregated totals per population across time
    compartmentPlot -- plots number of naive,inf,rec,dead hosts/vectors vs time
    compositionPlot -- plots counts for pathogen genomes or resistance vs. time

    --- Model interventions: ---

    - Make and connect populations -
    newPopulation -- create a new Population object with setup parameters
    linkPopulations -- set migration rate from one population towards another
    createInterconnectedPopulations -- create new populations, link all of them
        to each other

    - Modify population parameters -
    setSetup -- assigns a given set of parameters to this population

    - Manipulate hosts and vectors in population -
    newHostGroup -- returns a list of random (healthy or any) hosts
    newVectorGroup -- returns a list of random (healthy or any) vectors
    addHosts -- adds hosts to the population
    addVectors -- adds vectors to the population
    removeHosts -- removes hosts from the population
    removeVectors -- removes vectors from the population
    addPathogensToHosts -- adds pathogens with specified genomes to hosts
    addPathogensToVectors -- adds pathogens with specified genomes to vectors
    treatHosts -- removes infections susceptible to given treatment from hosts
    treatVectors -- removes infections susceptible to treatment from vectors
    protectHosts -- adds protection sequence to hosts
    protectVectors -- adds protection sequence to vectors


    --- Preset fitness functions: ---
        * these are static methods

    stabilizingSelection -- evaluates genome fitness by decreasing with distance
        from optimal sequence
    disruptiveSelection -- evaluates genome fitness by increasing with distance
        from worst sequence
    """

    ### CONSTANTS ###
    ### Color scheme constant ###
    CB_PALETTE = ["#E69F00", "#56B4E9", "#009E73",
                  "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#999999"]
     # www.cookbook-r.com/Graphs/Colors_(ggplot2)/#a-colorblind-friendly-palette
     # http://jfly.iam.u-tokyo.ac.jp/color/


    ### CLASS CONSTRUCTOR ###

    def __init__(self):
        """Create a new Model object."""
        super(Model, self).__init__()
        self.populations = {}
            # dictionary with keys=population IDs, values=Population objects
        self.setups = {}
            # dictionary with keys=setup IDs, values=Setup objects
        self.interventions = []
            # contains model interventions in the order they will occur
        self.groups = {}
            # dictionary with keys=group IDs, values=lists of hosts/vectors
        self.history = {}
            # dictionary with keys=time values, values=Model objects that are
            # snapshots of Model at that timepoint


    ### MODEL METHODS ###

    ### Model initialization and simulation: ###

    def newSetup(
            self, name, preset=None,
            num_loci=None, possible_alleles=None,
            fitnessHost=None, fitnessVector=None,
            contact_rate_host_vector=None, contact_rate_host_host=None,
            mean_inoculum_host=None, mean_inoculum_vector=None,
            recovery_rate_host=None, recovery_rate_vector=None,
            recombine_in_host=None, recombine_in_vector=None,
            mutate_in_host=None, mutate_in_vector=None,
            death_rate_host=None, death_rate_vector=None,
            protection_upon_recovery_host=None,
            protection_upon_recovery_vector=None):
        """Create a new Setup, save it in setups dict under given name.

        Two preset setups exist: "vector-borne" and "host-host". You may select
        one of the preset setups with the preset keyword argument and then
        modify individual parameters with additional keyword arguments, without
        having to specify all of them.

        Preset parameter setups:

        "vector-borne":
        num_loci = num_loci or 10
        possible_alleles = possible_alleles or 'ATCG'
        fitnessHost = fitnessHost or (lambda g: 1)
        fitnessVector = fitnessVector or (lambda g: 1)
        contact_rate_host_vector = contact_rate_host_vector or 1e1
        contact_rate_host_host = contact_rate_host_host or 0
        mean_inoculum_host = mean_inoculum_host or 1e2
        mean_inoculum_vector = mean_inoculum_vector or 1e2
        recovery_rate_host = recovery_rate_host or 1e-1
        recovery_rate_vector = recovery_rate_vector or 1e-2
        recombine_in_host = recombine_in_host or 0
        recombine_in_vector = recombine_in_vector or 1e-2
        mutate_in_host = mutate_in_host or 1e-6
        mutate_in_vector = mutate_in_vector or 0
        death_rate_host = death_rate_host or 0
        death_rate_vector = death_rate_vector or 0
        protection_upon_recovery_host = ( protection_upon_recovery_host
            or None )
        protection_upon_recovery_vector = ( protection_upon_recovery_vector
            or None )

        "host-host":
        num_loci = num_loci or 10
        possible_alleles = possible_alleles or 'ATCG'
        fitnessHost = fitnessHost or (lambda g: 1)
        fitnessVector = fitnessVector or (lambda g: 1)
        contact_rate_host_vector = contact_rate_host_vector or 0
        contact_rate_host_host = contact_rate_host_host or 2e1
        mean_inoculum_host = mean_inoculum_host or 1e1
        mean_inoculum_vector = mean_inoculum_vector or 0
        recovery_rate_host = recovery_rate_host or 1e-1
        recovery_rate_vector = recovery_rate_vector or 1e1
        recombine_in_host = recombine_in_host or 1e-3
        recombine_in_vector = recombine_in_vector or 0
        mutate_in_host = mutate_in_host or 1e-6
        mutate_in_vector = mutate_in_vector or 0
        death_rate_host = death_rate_host or 0
        death_rate_vector = death_rate_vector or 0
        protection_upon_recovery_host = ( protection_upon_recovery_host
            or None )
        protection_upon_recovery_vector = ( protection_upon_recovery_vector
            or None )

        Arguments:
        name -- name of setup to be used as a key in model setups dictionary

        Keyword arguments:
        preset -- preset setup to be used: "vector-borne" or "host-host", if
            None, must define all other keyword arguments (default None; None or
            String)
        num_loci -- length of each pathogen genome string (int > 0)
        possible_alleles -- set of possible characters in all genome string, or
            at each position in genome string (String or list of Strings with
            num_loci elements)
        fitnessHost -- relative fitness in head-to-head competition within host
            (number >= 0)
        fitnessVector -- relative fitness in head-to-head competition within
            vector (number >= 0)
        contact_rate_host_vector -- rate of host-vector contact events, not
            necessarily transmission, assumes constant population density;
            evts/time (number >= 0)
        contact_rate_host_host -- rate of host-host contact events, not
            necessarily transmission, assumes constant population density;
            evts/time (number >= 0)
        mean_inoculum_host -- mean number of pathogens that are transmitted from
            a vector or host into a new host during a contact event (int >= 0)
        mean_inoculum_vector -- mean number of pathogens that are transmitted
            from a host to a vector during a contact event (int >= 0)
        recovery_rate_host -- rate at which hosts clear all pathogens;
            1/time (number >= 0)
        recovery_rate_vector -- rate at which vectors clear all pathogens
            1/time (number >= 0)
        recombine_in_host -- rate at which recombination occurs in host;
            evts/time (number >= 0)
        recombine_in_vector -- rate at which recombination occurs in vector;
            evts/time (number >= 0)
        mutate_in_host -- rate at which mutation occurs in host; evts/time
            (number >= 0)
        mutate_in_vector -- rate at which mutation occurs in vector; evts/time
            (number >= 0)
        death_rate_host -- infected host death rate; 1/time (number >= 0)
        death_rate_vector -- infected vector death rate; 1/time (number >= 0)
        protection_upon_recovery_host -- defines indexes in genome string that
            define substring to be added to host protection sequences after
            recovery (None or array-like of length 2 with int 0-num_loci)
        protection_upon_recovery_vector -- defines indexes in genome string that
            define substring to be added to vector protection sequences after
            recovery (None or array-like of length 2 with int 0-num_loci)
        """

        if preset == "vector-borne":
            num_loci = num_loci or 10
            possible_alleles = possible_alleles or 'ATCG'
            fitnessHost = fitnessHost or (lambda g: 1)
            fitnessVector = fitnessVector or (lambda g: 1)
            contact_rate_host_vector = contact_rate_host_vector or 1e1
            contact_rate_host_host = contact_rate_host_host or 0
            mean_inoculum_host = mean_inoculum_host or 1e2
            mean_inoculum_vector = mean_inoculum_vector or 1e2
            recovery_rate_host = recovery_rate_host or 1e-1
            recovery_rate_vector = recovery_rate_vector or 1e-2
            recombine_in_host = recombine_in_host or 0
            recombine_in_vector = recombine_in_vector or 1e-2
            mutate_in_host = mutate_in_host or 1e-6
            mutate_in_vector = mutate_in_vector or 0
            death_rate_host = death_rate_host or 0
            death_rate_vector = death_rate_vector or 0
            protection_upon_recovery_host = ( protection_upon_recovery_host
                or None )
            protection_upon_recovery_vector = ( protection_upon_recovery_vector
                or None )

        elif preset == "host-host":
            num_loci = num_loci or 10
            possible_alleles = possible_alleles or 'ATCG'
            fitnessHost = fitnessHost or (lambda g: 1)
            fitnessVector = fitnessVector or (lambda g: 1)
            contact_rate_host_vector = contact_rate_host_vector or 0
            contact_rate_host_host = contact_rate_host_host or 2e1
            mean_inoculum_host = mean_inoculum_host or 1e1
            mean_inoculum_vector = mean_inoculum_vector or 0
            recovery_rate_host = recovery_rate_host or 1e-1
            recovery_rate_vector = recovery_rate_vector or 1e1
            recombine_in_host = recombine_in_host or 1e-3
            recombine_in_vector = recombine_in_vector or 0
            mutate_in_host = mutate_in_host or 1e-6
            mutate_in_vector = mutate_in_vector or 0
            death_rate_host = death_rate_host or 0
            death_rate_vector = death_rate_vector or 0
            protection_upon_recovery_host = ( protection_upon_recovery_host
                or None )
            protection_upon_recovery_vector = ( protection_upon_recovery_vector
                or None )

        self.setups[name] = Setup(
            num_loci, possible_alleles,
            fitnessHost, fitnessVector,
            contact_rate_host_vector, contact_rate_host_host,
            mean_inoculum_host, mean_inoculum_vector,
            recovery_rate_host, recovery_rate_vector,
            recombine_in_host, recombine_in_vector,
            mutate_in_host, mutate_in_vector,
            death_rate_host, death_rate_vector,
            protection_upon_recovery_host, protection_upon_recovery_vector
            )

    def newIntervention(self, time, function, args):
        """Create a new intervention to be carried out at a specific time.

        Arguments:
        time -- time at which intervention will take place (number)
        function -- intervention to be carried out (method of class Model)
        args -- contains arguments for function in positinal order (array-like)
        """

        self.interventions.append( Intervention(time, function, args) )

    def run(self,t0,tf):
        """Simulate model for a specified time between two time points.

        Simulates a time series using the Gillespie algorithm.

        Saves a dictionary containing model state history, with keys=times and
        values=Model objects with model snapshot at that time point under this
        model's history attribute.

        Arguments:
        t0 -- initial time point to start simulation at (number)
        tf -- initial time point to end simulation at (number)
        """

        sim = Gillespie(self)
        self.history = sim.run(t0,tf)


    ### Output and Plots: ###

    def saveToDataFrame(self,save_to_file,n_cores=0):
        """Save status of model to dataframe, write to file location given.

        Creates a pandas Dataframe in long format with the given model history,
        with one host or vector per simulation time in each row, and columns:
            Time - simulation time of entry
            Population - ID of this host/vector's population
            Organism - host/vector
            ID - ID of host/vector
            Pathogens - all genomes present in this host/vector separated by ;
            Protection - all genomes present in this host/vector separated by ;
            Alive - whether host/vector is alive at this time, True/False

        Arguments:
        save_to_file -- file path and name to save model data under (String)

        Keyword arguments:
        n_cores -- number of cores to parallelize file export across, if 0, all
            cores available are used (default 0; int)

        Returns:
        pandas dataframe with model history as described above
        """

        data = saveToDf(self.history,save_to_file,n_cores)

        return data

    def getPathogens(self, dat, save_to_file=""):
        """Create Dataframe with counts for all pathogen genomes in data.

        Returns sorted pandas Dataframe with counts for occurrences of all pathogen
        genomes in data passed.

        Arguments:
        data -- dataframe with model history as produced by saveToDf function

        Keyword arguments:
        save_to_file -- file path and name to save model data under, no saving
            occurs if empty string (default ''; String)

        Returns:
        pandas dataframe with Series as described above
        """

        return getPathogens(dat, save_to_file=save_to_file)

    def getProtections(self, dat, save_to_file=""):
        """Create Dataframe with counts for all protection sequences in data.

        Returns sorted pandas Dataframe with counts for occurrences of all
        protection sequences in data passed.

        Arguments:
        data -- dataframe with model history as produced by saveToDf function

        Keyword arguments:
        save_to_file -- file path and name to save model data under, no saving
            occurs if empty string (default ''; String)

        Returns:
        pandas dataframe with Series as described above
        """

        return getProtections(dat, save_to_file=save_to_file)

    def populationsPlot(
            self, file_name, data, compartment='Infected',
            hosts=True, vectors=False, num_top_populations=7,
            track_specific_populations=[], save_data_to_file="",
            x_label='Time', y_label='Hosts', figsize=(8, 4), dpi=200,
            palette=CB_PALETTE, stacked=False):
        """Create plot with aggregated totals per population across time.

        Creates a line or stacked line plot with dynamics of a compartment
        across populations in the model, with one line for each population.

        A host or vector is considered part of the recovered compartment
        if it has protection sequences of any kind and is not infected.

        Arguments:
        file_name -- file path, name, and extension to save plot under (String)
        data -- dataframe with model history as produced by saveToDf function
            (DataFrame)

        Keyword arguments:
        compartment -- subset of hosts/vectors to count totals of, can be either
            'Naive','Infected','Recovered', or 'Dead' (default 'Infected';
            String)
        hosts -- whether to count hosts (default True, Boolean)
        vectors -- whether to count vectors (default False, Boolean)
        num_top_populations -- how many populations to count separately and
            include as columns, remainder will be counted under column "Other";
             if <0, includes all populations in model (default 7; int)
        track_specific_populations -- contains IDs of specific populations to
            have as a separate column if not part of the top num_top_populations
            populations (list of Strings)
        save_data_to_file -- file path and name to save model plot data under,
            no saving occurs if empty string (default ''; String)
        x_label -- X axis title (default 'Time', String)
        y_label -- Y axis title (default 'Hosts', String)
        legend_title -- legend title (default 'Population', String)
        legend_values -- labels for each trace, if empty list, uses population
            IDs (default empty list, list of Strings)
        figsize -- dimensions of figure (default (8,4), array-like of two ints)
        dpi -- figure resolution (default 200, int)
        palette -- color palette to use for traces (default CB_PALETTE, list of
            color Strings)
        stacked -- whether to draw a regular line plot or a stacked one (default
            False, Boolean)

        Returns:
        axis object for plot with model population dynamics as described above
        """

        return populationsPlot(
            file_name, data, compartment=compartment, hosts=hosts,
            vectors=vectors, num_top_populations=num_top_populations,
            track_specific_populations=track_specific_populations,
            save_data_to_file=save_data_to_file,
            x_label=x_label, y_label=y_label, figsize=figsize, dpi=dpi,
            palette=palette, stacked=stacked
            )

    def compartmentPlot(
            self, file_name, data, populations=[], hosts=True, vectors=False,
            save_data_to_file="", x_label='Time', y_label='Hosts',
            figsize=(8, 4), dpi=200, palette=CB_PALETTE, stacked=False):
        """Create plot with number of naive,inf,rec,dead hosts/vectors vs. time.

        Creates a line or stacked line plot with dynamics of all compartments
        (naive, infected, recovered, dead) across selected populations in the
        model, with one line for each compartment.

        A host or vector is considered part of the recovered compartment
        if it has protection sequences of any kind and is not infected.

        Arguments:
        file_name -- file path, name, and extension to save plot under (String)
        data -- dataframe with model history as produced by saveToDf function
            (DataFrame)

        Keyword arguments:
        populations -- IDs of populations to include in analysis; if empty, uses
            all populations in model (default empty list; list of Strings)
        hosts -- whether to count hosts (default True, Boolean)
        vectors -- whether to count vectors (default False, Boolean)
        save_to_file -- file path and name to save model data under, no saving
            occurs if empty string (default ''; String)
        x_label -- X axis title (default 'Time', String)
        y_label -- Y axis title (default 'Hosts', String)
        legend_title -- legend title (default 'Population', String)
        legend_values -- labels for each trace, if empty list, uses population
            IDs (default empty list, list of Strings)
        figsize -- dimensions of figure (default (8,4), array-like of two ints)
        dpi -- figure resolution (default 200, int)
        palette -- color palette to use for traces (default CB_PALETTE, list of
            color Strings)
        stacked -- whether to draw a regular line plot or a stacked one (default
            False, Boolean)

        Returns:
        axis object for plot with model compartment dynamics as described above
        """

        return compartmentPlot(
            file_name, data, populations=populations, hosts=hosts,
            vectors=vectors, save_data_to_file=save_data_to_file,
            x_label=x_label, y_label=y_label, figsize=figsize, dpi=dpi,
            palette=palette, stacked=stacked
            )

    def compositionPlot(
            self, file_name, data, populations=[],
            type_of_composition='Pathogens', hosts=True, vectors=False,
            num_top_sequences=7, track_specific_genomes=[],
            save_data_to_file="", x_label='Time', y_label='Infections',
            figsize=(8, 4), dpi=200, palette=CB_PALETTE, stacked=True):
        """Create plot with counts for pathogen genomes or resistance vs. time.

        Creates a line or stacked line plot with dynamics of the pathogen
        strains or protection sequences across selected populations in the
        model, with one line for each pathogen genome or protection sequence
        being shown.

        Of note: sum of totals for all sequences in one time point does not
        necessarily equal the number of infected hosts and/or vectors, given
        multiple infections in the same host/vector are counted separately.

        Arguments:
        data -- dataframe with model history as produced by saveToDf function

        Keyword arguments:
        populations -- IDs of populations to include in analysis; if empty, uses
            all populations in model (default empty list; list of Strings)
        type_of_composition -- field of data to count totals of, can be either
            'Pathogens' or 'Protection' (default 'Pathogens'; String)
        hosts -- whether to count hosts (default True, Boolean)
        vectors -- whether to count vectors (default False, Boolean)
        num_top_sequences -- how many sequences to count separately and include
            as columns, remainder will be counted under column "Other"; if <0,
            includes all genomes in model (default 7; int)
        track_specific_genomes -- contains specific sequences to have
            as a separate column if not part of the top num_top_sequences
            sequences (list of Strings)
        save_to_file -- file path and name to save model data under, no saving
            occurs if empty string (default ''; String)
        x_label -- X axis title (default 'Time', String)
        y_label -- Y axis title (default 'Hosts', String)
        legend_title -- legend title (default 'Population', String)
        legend_values -- labels for each trace, if empty list, uses population
            IDs (default empty list, list of Strings)
        figsize -- dimensions of figure (default (8,4), array-like of two ints)
        dpi -- figure resolution (default 200, int)
        palette -- color palette to use for traces (default CB_PALETTE, list of
            color Strings)
        stacked -- whether to draw a regular line plot or a stacked one (default
            False, Boolean)

        Returns:
        axis object for plot with model sequence composition dynamics as
            described
        """

        return compositionPlot(
            file_name, data, populations=populations,
            type_of_composition=type_of_composition, hosts=hosts,
            vectors=vectors, num_top_sequences=num_top_sequences,
            track_specific_genomes=track_specific_genomes,
            save_data_to_file=save_data_to_file,
            x_label=x_label, y_label=y_label, figsize=figsize, dpi=dpi,
            palette=palette, stacked=stacked
            )

    ### Model interventions: ###

    def newPopulation(self, id, setup_name, num_hosts=100, num_vectors=100):
        """Create a new Population object with setup parameters.

        If population ID is already in use, appends _2 to it

        Arguments:
        id -- unique identifier for this population in the model (String)
        setup_name -- setup object with parameters for this population (Setup)

        Keyword arguments:
        num_hosts -- number of hosts to initialize population with (default 100;
            int)
        num_vectors -- number of hosts to initialize population with (default
            100; int)
        """

        if id in self.populations.keys():
            id = id+'_2'

        self.populations[id] = Population(
            self, id, self.setups[setup_name], num_hosts, num_vectors
            )

    def linkPopulations(self, pop1_id, pop2_id, rate):
        """Set migration rate from one population towards another.

        Arguments:
        neighbor -- population towards which migration rate will be specified
            (Population)
        rate -- migration rate from this population to the neighbor; evts/time
            (number)
        """

        self.populations[pop1_id].setNeighbor( self.populations[pop2_id], rate )

    def createInterconnectedPopulations(
            self, num_populations, migration_rate, id_prefix, setup_name,
            num_hosts=100, num_vectors=100):
        """Create new populations, link all of them to each other.

        All populations in this cluster are linked with the same migration rate,
        starting number of hosts and vectors, and setup parameters. Their IDs
        are numbered onto prefix given as 'id_prefix_0', 'id_prefix_1',
        'id_prefix_2', etc.

        Arguments:
        num_populations -- number of populations to be created (int)
        migration_rate -- migration rate between populations; evts/time (number)
        id_prefix -- prefix for IDs to be used for this population in the model,
            (String)
        setup_name -- setup object with parameters for all populations (Setup)

        Keyword arguments:
        num_hosts -- number of hosts to initialize population with (default 100;
            int)
        num_vectors -- number of hosts to initialize population with (default
            100; int)
        """

        new_pops = [
            Population(
                self, id_prefix + str(i), self.setups[setup_name],
                num_hosts, num_vectors
                ) for i in range(num_populations)
            ]
        new_pop_ids = []
        for pop in new_pops:
            if pop.id in self.populations.keys():
                pop.id = pop.id+'_2'

            self.populations[pop.id] = pop
            new_pop_ids.append(pop.id)

        for p1_id in new_pop_ids:
            for p2_id in new_pop_ids:
                self.linkPopulations(p1_id,p2_id,migration_rate)

    def newHostGroup(self, pop_id, group_id, num_hosts, healthy=False):
        """Return a list of random (healthy or any) hosts in population.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_vectors -- number of vectors to be sampled randomly (int)

        Keyword arguments:
        healthy -- whether to sample healthy hosts only (default True; Boolean)

        Returns:
        list containing sampled hosts
        """

        self.groups[group_id] = self.populations[pop_id].newHostGroup(num_hosts)

    def newVectorGroup(self, pop_id, group_id, num_vectors, healthy=False):
        """Return a list of random (healthy or any) vectors in population.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_vectors -- number of vectors to be sampled randomly (int)

        Keyword arguments:
        healthy -- whether to sample healthy vectors only (default True;
            Boolean)

        Returns:
        list containing sampled vectors
        """

        self.groups[group_id] = self.populations[pop_id].newVectorGroup(
            num_vectors
            )

    def addHosts(self, pop_id, num_hosts):
        """Add a number of healthy hosts to population, return list with them.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_hosts -- number of hosts to be added (int)

        Returns:
        list containing new hosts
        """

        self.populations[pop_id].addHosts(num_hosts)

    def addVectors(self, pop_id, num_vectors):
        """Add a number of healthy vectors to population, return list with them.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_vectors -- number of vectors to be added (int)

        Returns:
        list containing new vectors
        """

        self.populations[pop_id].addVectors(num_vectors)

    def removeHosts(self, pop_id, num_hosts_or_list):
        """Remove a number of specified or random hosts from population.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_hosts_or_list -- number of hosts to be sampled randomly for removal
            or list of hosts to be removed, must be hosts in this population
            (int or list of Hosts)
        """

        self.populations[pop_id].removeHosts(num_hosts_or_list)

    def removeVectors(self, pop_id, num_vectors_or_list):
        """Remove a number of specified or random vectors from population.

        Arguments:
        pop_id -- ID of population to be modified (String)
        num_vectors_or_list -- number of vectors to be sampled randomly for
            removal or list of vectors to be removed, must be vectors in this
            population (int or list of Vectors)
        """

        self.populations[pop_id].removeVectors(num_vectors_or_list)


    def addPathogensToHosts(self, pop_id, genomes_numbers, group_id=""):
        """Add specified pathogens to random hosts, optionally from a list.

        Arguments:
        pop_id -- ID of population to be modified (String)
        genomes_numbers -- dictionary containing pathogen genomes to add as keys
            and number of hosts each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        if group_id == "":
            hosts = self.populations[pop_id].hosts
        else:
            hosts = self.groups[group_id]

        self.populations[pop_id].addPathogensToHosts(genomes_numbers,hosts)

    def addPathogensToVectors(self, pop_id, genomes_numbers, group_id=""):
        """Add specified pathogens to random vectors, optionally from a list.

        Arguments:
        pop_id -- ID of population to be modified (String)
        genomes_numbers -- dictionary containing pathogen genomes to add as keys
            and number of vectors each one will be added to as values (dict with
            keys=Strings, values=int)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        if group_id == "":
            vectors = self.populations[pop_id].vectors
        else:
            vectors = self.groups[group_id]

        self.populations[pop_id].addPathogensToVectors(genomes_numbers,vectors)

    def treatHosts(self, pop_id, frac_hosts, resistance_seqs, group_id=""):
        """Treat random fraction of infected hosts against some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        pop_id -- ID of population to be modified (String)
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        if group_id == "":
            hosts = self.populations[pop_id].hosts
        else:
            hosts = self.groups[group_id]

        self.populations[pop_id].treatHosts(frac_hosts,resistance_seqs,hosts)

    def treatVectors(self, pop_id, frac_vectors, resistance_seqs, group_id=""):
        """Treat random fraction of infected vectors agains some infection.

        Removes all infections with genotypes susceptible to given treatment.
        Pathogens are removed if they are missing at least one of the sequences
        in resistance_seqs from their genome. Removes this organism from
        population infected list and adds to healthy list if appropriate.

        Arguments:
        pop_id -- ID of population to be modified (String)
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        resistance_seqs -- contains sequences required for treatment resistance
            (list of Strings)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        if group_id == "":
            vectors = self.populations[pop_id].vectors
        else:
            vectors = self.groups[group_id]

        self.populations[pop_id].treatVectors(
            frac_vectors,resistance_seqs,vectors
            )

    def protectHosts(
            self, pop_id, frac_hosts, protection_sequence, group_id=""):
        """Protect a random fraction of infected hosts against some infection.

        Adds protection sequence specified to a random fraction of the hosts
        specified. Does not cure them if they are already infected.

        Arguments:
        pop_id -- ID of population to be modified (String)
        frac_hosts -- fraction of hosts considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        hosts -- list of specific hosts to sample from, if empty, samples from
            whole population (default empty list; empty)
        """

        if group_id == "":
            hosts = self.populations[pop_id].hosts
        else:
            hosts = self.groups[group_id]

        self.populations[pop_id].protectHosts(
            frac_hosts,protection_sequence,hosts
            )

    def protectVectors(
            self, pop_id, frac_vectors, protection_sequence, group_id=""):
        """Protect a random fraction of infected vectors against some infection.

        Adds protection sequence specified to a random fraction of the vectors
        specified. Does not cure them if they are already infected.

        Arguments:
        pop_id -- ID of population to be modified (String)
        frac_vectors -- fraction of vectors considered to be randomly selected
            (number between 0 and 1)
        protection_sequence -- sequence against which to protect (String)

        Keyword arguments:
        vectors -- list of specific vectors to sample from, if empty, samples
            from whole population (default empty list; empty)
        """

        if group_id == "":
            vectors = self.populations[pop_id].vectors
        else:
            vectors = self.groups[group_id]

        self.populations[pop_id].protectVectors(
            frac_vectors,protection_sequence,vectors
            )

    def setSetup(self, pop_id, setup_id):
        """Assign parameters stored in Setup object to this population.

        Arguments:
        pop_id -- ID of population to be modified (String)
        setup_id -- ID of setup to be assigned (String)
        """

        self.populations[pop_id].setSetup( self.setups[setup_id] )


    ### Preset fitness functions: ###

    @staticmethod
    def stabilizingSelection(genome, optimal_genome, min_fitness):
        """Evaluate genome fitness by decreasing with distance from optimal seq.

        A purifying selection fitness function based on exponential decay of
        fitness as genomes move away from the optimal sequence. Distance is
        measured as Hamming distance from an optimal genome sequence.

        Arguments:
        genome -- the genome to be evalued (String)
        optimal_genome -- the genome sequence to measure distance against, has
            fitness of 1 (String)
        min_fitness -- minimum fitness value at maximum distance from optimal
            genome (number > 0)

        Return:
        fitness value of genome (number)
        """

        similarity = dl.SequenceMatcher(None, genome, optimal_genome).ratio()
        fitness = np.exp( np.log( min_fitness ) * ( 1-similarity ) )

        return fitness

    @staticmethod
    def disruptiveSelection(genome, worst_genome, min_fitness):
        """Evaluate genome fitness by increasing with distance from worst seq.

        A purifying selection fitness function based on exponential decay of
        fitness as genomes move closer to the worst possible sequence. Distance
        is measured as Hamming distance from the worst possible genome sequence.

        Arguments:
        genome -- the genome to be evalued (String)
        optimal_genome -- the genome sequence to measure distance against, has
            fitness of min_fitness (String)
        min_fitness -- fitness value of worst possible genome (number > 0)

        Return:
        fitness value of genome (number)
        """

        similarity = dl.SequenceMatcher(None, genome, optimal_genome).ratio()
        fitness = np.exp( np.log( min_fitness ) * ( similarity ) )

        return fitness





"""Contains graphmaking methods."""

### Imports ###
import matplotlib.pyplot as plt # plots
import seaborn as sns # pretty plots
import numpy as np # handle arrays
import pandas as pd # data wrangling
import copy as cp

from opqua.internal.data import saveToDf, populationsDf, compartmentDf, \
    compositionDf

CB_PALETTE = ["#E69F00", "#56B4E9", "#009E73",
              "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#999999"]
    # www.cookbook-r.com/Graphs/Colors_(ggplot2)/#a-colorblind-friendly-palette
    # http://jfly.iam.u-tokyo.ac.jp/color/


def populationsPlot(
        file_name, data, compartment='Infected', hosts=True, vectors=False,
        num_top_populations=7, track_specific_populations=[],
        save_data_to_file="", x_label='Time', y_label='Infected hosts',
        legend_title='Population', legend_values=[], figsize=(8, 4), dpi=200,
        palette=CB_PALETTE, stacked=False):
    """Create plot with aggregated totals per population across time.

    Creates a line or stacked line plot with dynamics of a compartment
    across populations in the model, with one line for each population.

    Arguments:
    file_name -- file path, name, and extension to save plot under (String)
    data -- dataframe with model history as produced by saveToDf function
        (DataFrame)

    Keyword arguments:
    compartment -- subset of hosts/vectors to count totals of, can be either
        'Naive','Infected','Recovered', or 'Dead' (default 'Infected'; String)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    num_top_populations -- how many populations to count separately and include
        as columns, remainder will be counted under column "Other"; if <0,
        includes all populations in model (default 7; int)
    track_specific_populations -- contains IDs of specific populations to have
        as a separate column if not part of the top num_top_populations
        populations (list of Strings)
    save_data_to_file -- file path and name to save model plot data under, no
        saving occurs if empty string (default ''; String)
    x_label -- X axis title (default 'Time', String)
    y_label -- Y axis title (default 'Hosts', String)
    legend_title -- legend title (default 'Population', String)
    legend_values -- labels for each trace, if empty list, uses population IDs
        (default empty list, list of Strings)
    figsize -- dimensions of figure (default (8,4), array-like of two ints)
    dpi -- figure resolution (default 200, int)
    palette -- color palette to use for traces (default CB_PALETTE, list of
        color Strings)
    stacked -- whether to draw a regular line plot or a stacked one (default
        False, Boolean)

    Returns:
    axis object for plot with model population dynamics as described above
    """

    pops = populationsDf(
        data, compartment=compartment, hosts=hosts, vectors=vectors,
        num_top_populations=num_top_populations,
        track_specific_populations=track_specific_populations,
        save_to_file=save_data_to_file
        )

    if pops.shape[1] > 0:
        plt.figure(figsize=figsize, dpi=dpi) # make new figure
        ax = plt.subplot(1, 1, 1) # get axis

        if stacked:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = pops.drop(columns='Time').columns

            ax.stackplot(
                pops['Time'], pops.drop(columns='Time').transpose(),
                labels=labs, colors=palette
                )
        else:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = pops.drop(columns='Time').columns

            for i,c in enumerate(pops.columns[1:]):
                ax.plot(
                    pops['Time'], pops[c], label=labs[i], color=palette[i]
                    )

        plt.xlabel(x_label) # labels
        plt.ylabel(y_label)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        handles, labels = ax.get_legend_handles_labels() # get legend
        plt.legend(
            handles, labels, loc='center left', bbox_to_anchor=(1, 0.5),
            title=legend_title
            ) # show it

        plt.savefig(file_name, bbox_inches='tight') # save
    else:
        ax = None
        print('Nothing to plot! Check your data.')

    return ax


def compartmentPlot(
        file_name, data, populations=[], hosts=True, vectors=False,
        save_data_to_file="", x_label='Time', y_label='Hosts',
        legend_title='Compartment', legend_values=[], figsize=(8, 4), dpi=200,
        palette=CB_PALETTE, stacked=False):
    """Create plot with num. of naive, susc., inf., rec. hosts/vectors vs. time.

    Creates a line or stacked line plot with dynamics of all compartments
    (naive, infected, recovered, dead) across selected populations in the model,
    with one line for each compartment.

    Arguments:
    file_name -- file path, name, and extension to save plot under (String)
    data -- dataframe with model history as produced by saveToDf function
        (DataFrame)

    Keyword arguments:
    populations -- IDs of populations to include in analysis; if empty, uses all
        populations in model (default empty list; list of Strings)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)
    x_label -- X axis title (default 'Time', String)
    y_label -- Y axis title (default 'Hosts', String)
    legend_title -- legend title (default 'Population', String)
    legend_values -- labels for each trace, if empty list, uses population IDs
        (default empty list, list of Strings)
    figsize -- dimensions of figure (default (8,4), array-like of two ints)
    dpi -- figure resolution (default 200, int)
    palette -- color palette to use for traces (default CB_PALETTE, list of
        color Strings)
    stacked -- whether to draw a regular line plot or a stacked one (default
        False, Boolean)

    Returns:
    axis object for plot with model compartment dynamics as described above
    """

    comp = compartmentDf(data, populations=populations, hosts=hosts,
        vectors=vectors, save_to_file=save_data_to_file)

    if comp.shape[1] > 0:
        plt.figure(figsize=figsize, dpi=dpi) # make new figure
        ax = plt.subplot(1, 1, 1) # get axis

        if stacked:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = comp.drop(columns='Time').columns

            ax.stackplot(
                comp['Time'], comp.drop(columns='Time').transpose(),
                labels=labs, colors=palette
                )
        else:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = comp.drop(columns='Time').columns

            for i,c in enumerate(comp.columns[1:]):
                ax.plot(
                    comp['Time'], comp[c], label=labs[i], color=palette[i]
                    )

        plt.xlabel(x_label) # labels
        plt.ylabel(y_label)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        handles, labels = ax.get_legend_handles_labels() # get legend
        plt.legend(
            handles, labels, loc='center left', bbox_to_anchor=(1, 0.5),
            title=legend_title
            ) # show it

        plt.savefig(file_name, bbox_inches='tight') # save
    else:
        ax = None
        print('Nothing to plot! Check your data.')

    return ax

def compositionPlot(
        file_name, data, populations=[], type_of_composition='Pathogens',
        hosts=True, vectors=False, num_top_sequences=7,
        track_specific_genomes=[], save_data_to_file="", x_label='Time',
        y_label='Infections', legend_title='Genotype', legend_values=[],
        figsize=(8, 4), dpi=200, palette=CB_PALETTE, stacked=True):
    """Create plot with counts for pathogen genomes or resistance across time.

    Creates a line or stacked line plot with dynamics of the pathogen strains or
    protection sequences across selected populations in the model,
    with one line for each pathogen genome or protection sequence being shown.

    Of note: sum of totals for all sequences in one time point does not
    necessarily equal the number of infected hosts and/or vectors, given
    multiple infections in the same host/vector are counted separately.

    Arguments:
    data -- dataframe with model history as produced by saveToDf function

    Keyword arguments:
    populations -- IDs of populations to include in analysis; if empty, uses all
        populations in model (default empty list; list of Strings)
    type_of_composition -- field of data to count totals of, can be either
        'Pathogens' or 'Protection' (default 'Pathogens'; String)
    hosts -- whether to count hosts (default True, Boolean)
    vectors -- whether to count vectors (default False, Boolean)
    num_top_sequences -- how many sequences to count separately and include
        as columns, remainder will be counted under column "Other"; if <0,
        includes all genomes in model (default 7; int)
    track_specific_genomes -- contains specific sequences to have
        as a separate column if not part of the top num_top_sequences
        sequences (list of Strings)
    save_to_file -- file path and name to save model data under, no saving
        occurs if empty string (default ''; String)
    x_label -- X axis title (default 'Time', String)
    y_label -- Y axis title (default 'Hosts', String)
    legend_title -- legend title (default 'Population', String)
    legend_values -- labels for each trace, if empty list, uses population IDs
        (default empty list, list of Strings)
    figsize -- dimensions of figure (default (8,4), array-like of two ints)
    dpi -- figure resolution (default 200, int)
    palette -- color palette to use for traces (default CB_PALETTE, list of
        color Strings)
    stacked -- whether to draw a regular line plot or a stacked one (default
        False, Boolean)

    Returns:
    axis object for plot with model sequence composition dynamics as described
    """

    comp = compositionDf(
        data, populations=populations, type_of_composition=type_of_composition,
        hosts=hosts, vectors=vectors, num_top_sequences=num_top_sequences,
        track_specific_genomes=track_specific_genomes,
        save_to_file=save_data_to_file
        )

    if comp.shape[1] > 0:
        plt.figure(figsize=figsize, dpi=dpi) # make new figure
        ax = plt.subplot(1, 1, 1) # get axis

        if stacked:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = comp.drop(columns='Time').columns

            ax.stackplot(
                comp['Time'], comp.drop(columns='Time').transpose(),
                labels=labs, colors=palette
                )
        else:
            if len(legend_values) > 0:
                labs = legend_values
            else:
                labs = comp.drop(columns='Time').columns

            for i,c in enumerate(comp.columns[1:]):
                ax.plot(
                    comp['Time'], comp[c], label=labs[i], color=palette[i]
                    )

        plt.xlabel(x_label) # labels
        plt.ylabel(y_label)

        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        handles, labels = ax.get_legend_handles_labels() # get legend
        plt.legend(
            handles, labels, loc='center left', bbox_to_anchor=(1, 0.5),
            title=legend_title
            ) # show it

        plt.savefig(file_name, bbox_inches='tight') # save
    else:
        ax = None
        print('Nothing to plot! Check your data.')

    return ax



"""
Opqua.

An epidemiological modeling framework for population genetics and evolution.
"""

__version__ = "0.1.0"
__author__ = 'Pablo Cardenas'
__credits__ = 'Massachusetts Institute of Technology'

# init file to make this directory into a python package
from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]




"""Contains class Intervention."""

class Setup(object):
    """Class defines a setup with population parameters."""

    def __init__(
            self,
            num_loci, possible_alleles,
            fitnessHost, fitnessVector,
            contact_rate_host_vector, contact_rate_host_host,
            mean_inoculum_host, mean_inoculum_vector,
            recovery_rate_host, recovery_rate_vector,
            recombine_in_host, recombine_in_vector,
            mutate_in_host, mutate_in_vector, death_rate_host,death_rate_vector,
            protection_upon_recovery_host, protection_upon_recovery_vector):
        """Create a new Setup.

        Arguments:
        num_loci -- length of each pathogen genome string (int > 0)
        possible_alleles -- set of possible characters in all genome string, or
            at each position in genome string (String or list of Strings with
            num_loci elements)
        fitnessHost -- relative fitness in head-to-head competition within host
            (number >= 0)
        fitnessVector -- relative fitness in head-to-head competition within
            vector (number >= 0)
        contact_rate_host_vector -- rate of host-vector contact events, not
            necessarily transmission, assumes constant population density;
            evts/time (number >= 0)
        contact_rate_host_host -- rate of host-host contact events, not
            necessarily transmission, assumes constant population density;
            evts/time (number >= 0)
        mean_inoculum_host -- mean number of pathogens that are transmitted from
            a vector or host into a new host during a contact event (int >= 0)
        mean_inoculum_vector -- mean number of pathogens that are transmitted
            from a host to a vector during a contact event (int >= 0)
        recovery_rate_host -- rate at which hosts clear all pathogens;
            1/time (number >= 0)
        recovery_rate_vector -- rate at which vectors clear all pathogens
            1/time (number >= 0)
        recombine_in_host -- rate at which recombination occurs in host;
            evts/time (number >= 0)
        recombine_in_vector -- rate at which recombination occurs in vector;
            evts/time (number >= 0)
        mutate_in_host -- rate at which mutation occurs in host; evts/time
            (number >= 0)
        mutate_in_vector -- rate at which mutation occurs in vector; evts/time
            (number >= 0)
        death_rate_host -- infected host death rate; 1/time (number >= 0)
        death_rate_vector -- infected vector death rate; 1/time (number >= 0)
        protection_upon_recovery_host -- defines indexes in genome string that
            define substring to be added to host protection sequences after
            recovery (None or array-like of length 2 with int 0-num_loci)
        protection_upon_recovery_vector -- defines indexes in genome string that
            define substring to be added to vector protection sequences after
            recovery (None or array-like of length 2 with int 0-num_loci)
        """

        super(Setup, self).__init__()
        self.num_loci = num_loci
        if isinstance(possible_alleles, list):
            self.possible_alleles = possible_alleles
        else:
            self.possible_alleles = [possible_alleles] * self.num_loci
                # possible_alleles must be a list with all available alleles for
                # each position

        self.fitnessHost = fitnessHost
        self.fitnessVector = fitnessVector
        self.contact_rate_host_vector = contact_rate_host_vector
        self.contact_rate_host_host = contact_rate_host_host
            # contact rates assumes scaling area--large populations are equally
            # dense as small ones, so contact is constant with both host and
            # vector populations. If you don't want this to happen, modify the
            # population's contact rate accordingly.
            # Examines contacts between infected hosts and all hosts
        self.mean_inoculum_host = mean_inoculum_host
        self.mean_inoculum_vector = mean_inoculum_vector
        self.recovery_rate_host = recovery_rate_host
        self.recovery_rate_vector = recovery_rate_vector

        self.recombine_in_host = recombine_in_host
        self.recombine_in_vector = recombine_in_vector
        self.mutate_in_host = mutate_in_host
        self.mutate_in_vector = mutate_in_vector

        self.death_rate_host = death_rate_host
        self.death_rate_vector = death_rate_vector
        self.protection_upon_recovery_host = protection_upon_recovery_host
        self.protection_upon_recovery_vector = protection_upon_recovery_vector



# init file to make this directory into a python package
from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]




"""Contains class Intervention."""

class Intervention(object):
    """Class defines a new intervention to be done at a specified time.

    Methods:
    doIntervention -- executes intervention function with specified arguments
    """

    def __init__(self, time, function, args):
        """Create a new Intervention.

        Arguments:
        time -- time at which intervention will take place (number)
        function -- intervention to be carried out (method of class Model)
        args -- contains arguments for function in positinal order (array-like)
        """
        super(Intervention, self).__init__()
        self.time = time
        self.intervention = function
        self.args = args

    def doIntervention(self):
        """Execute intervention function with specified arguments."""

        self.intervention(*self.args)
